#!/bin/false
# -*- coding: utf-8 -*-

""" Object-oriented filesystem helper classes. """


# ------------------------------------------------------------------------------------------------------------------------
# Import and pathing setup
# ------------------------------------------------------------------------------------------------------------------------
import pathlib
from typing import Any, Callable, Dict, Generator, Iterable, Optional, Pattern, Union

from pt.ptutils.io import get_loader, get_saver, load_txt, save_txt
from pt.ptutils.globbing import get_subdirs, get_subfiles
from pt.ptutils.text import strip_line_ending
from pt.ptutils.undefined import UNDEFINED, is_defined


# ------------------------------------------------------------------------------------------------------------------------
# Typehint helpers
# ------------------------------------------------------------------------------------------------------------------------
""" Typehint for things which are like a filesystem path. """
PathLike = Union[str, pathlib.Path, 'Path', 'File', 'Folder']

""" Typehint for things which are like a list of filesystem paths. """
PathList = Iterable[PathLike]


# ------------------------------------------------------------------------------------------------------------------------
# Class: Path
# ------------------------------------------------------------------------------------------------------------------------
class Path(pathlib.Path):
    """ A filesystem path. """

    def __init__(self, path: PathLike):
        """
        Create a new filesystem path object.

        Parameters
        ----------
        path : PathLike
            The filesystem path. May be a `Path`, a `pathlib.Path`, a string, or any type deriving from `Path`.
        """
        super().__init__(path)

    # --------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def path_of(path: PathLike) -> 'Path':
        """
        Given a `PathLike` object, coerce it into being a `Path`.

        Returns
        -------
        Path
            A path object referencing the same path as `path`
        """
        return Path(path)

    # --------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def names_of(paths: PathList) -> Generator[str, None, None]:
        """
        Given a list of path objects, return a list of the filename portion of each path.

        Parameters
        ----------
        paths : PathList
            An iterable of `PathLike` objects. To determine filename, these will be coerced
            into `Path` objects before retrieving the filename portion of the path.

        Yields
        ------
        str
            The filename portion of each path provided in `paths`
        """
        for path in paths:
            yield Path.path_of(path).name

    # --------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def relative_paths_to(paths: PathList, reference: PathLike) -> Generator[str, None, None]:
        """
        Given a list of paths and a reference path, compute and return a list of the relative paths.

        Parameters
        ----------
        paths : PathList
            An iterable of `PathLike` objects. To relative path filename, these will be coerced
            into `Path` objects before retrieving the relative path with respect to `reference`.
        reference: PathLike
            A path from which to compute the relative paths of each path in `paths`. this will be
            coerced to a `Path` object before use.

        Yields
        ------
        str
            The filename portion of each path provided in `paths`
        """
        reference = Path.path_of(reference)
        for path in paths:
            yield pathlib.Path(path).relative_to(reference)


# =======================================================================================================================
# Class: File
# =======================================================================================================================
class File(Path):
    """ Class to simplify common file operations. """

    # --------------------------------------------------------------------------------------------------------------------
    def __init__(self, path: PathLike):
        """
        Create a new File path object.

        Parameters
        ----------
        path : PathLike
            The filesystem path of a file. This file may or may not exist.
        """
        super().__init__(path)
        self._loader  = None
        self._saver   = None

    # --------------------------------------------------------------------------------------------------------------------
    def exists(self) -> bool:
        """
        Test whether this path exists and refers to a file.

        Returns
        -------
        bool
            True if the path refers to an existing file.
        """
        return super().exists() and self.is_file()

    # --------------------------------------------------------------------------------------------------------------------
    def initialize_content(self, content: Any = UNDEFINED) -> None:
        """
        Initialize a file's contents by setting the content property if the file doesn't already exist.

        Parameters
        ----------
        content : Any, optional
            [description], by default UNDEFINED
        """
        if is_defined(content):
            if not self.exists():
                self.content = content

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def loader(self) -> Callable[[PathLike], Any]:
        """
        Get the most appropriate loader for this path based on the filename's extension.

        Returns
        -------
        Callable[[PathLike], Any]
            A function which loads the content of the file.

        """
        if self._loader is None:
            self._loader = get_loader(
                filename = str(self),
                default  = load_txt
            )
        return self._loader

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def saver(self) -> Callable[[PathLike, Any], None]:
        """
        Get the most appropriate saver for this path based on the filename's extension.

        Returns
        -------
        Callable[[PathLike, Any], None]
            A function which saves an object to the file in whatever format is most appropriate.

        """
        if self._saver is None:
            self._saver = get_saver(
                filename = str(self),
                default  = save_txt
            )
        return self._saver

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def content(self) -> Any:
        """
        Load the file's contents using the loader determined when the `File` object was created.

        Returns
        -------
        Any
            The loaded/decoded file content. The structure and format is determined by the
        """
        return self.loader(self)

    # --------------------------------------------------------------------------------------------------------------------
    @content.setter
    def content(self, value: Any) -> None:
        """
        Set the file's content by invoking the appropriate saver.

        Parameters
        ----------
        value : Any
            The value to encode and save to the file.
        """
        self.saver(self, value)

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def lines(self) -> Generator[str, None, None]:
        """
        Iterator over the text lines in a file.

        Yields
        ------
        str
            A line of text from the file.

        Notes
        -----
            This property bypasses the determined loader and instead directly
            reads the file as text. If used with binary files, undefined behaviour may result.
        """
        with open(self, 'r') as filp:
            for line in filp.readlines():
                yield strip_line_ending( line )


# =======================================================================================================================
# Class: Folder
# =======================================================================================================================
class Folder(Path):
    """ Class to simplify common folder operations. """

    # --------------------------------------------------------------------------------------------------------------------
    def exists(self) -> bool:
        """
        Test whether this path exists and refers to a folder.

        Returns
        -------
        bool
            True if the path refers to an existing folder.
        """
        return super().exists() and self.is_dir()

    # --------------------------------------------------------------------------------------------------------------------
    def search_subdirectories(
        self,
        pattern:   Optional[Pattern] = None,
        recursive: bool    = False
    ) -> Generator['Folder', None, None]:
        """
        Search for subdirectories of this folder, optionally recursive, optionally
        matching a regular expression.

        Parameters
        ----------
        pattern : Pattern, optional
            A regular expression to match against folder basenames, by default
            None. When omitted, every folder in the search folder will be returned.
        recursive : bool, optional
            When True, search recursively into all subfolders of the search
            folder, by default False.

        Yields
        -------
        Folder
            Any folders matching the constraints specified.
        """
        for fullpath in get_subdirs(self, pattern=pattern, recursive=recursive):
            yield Folder(fullpath)

    # --------------------------------------------------------------------------------------------------------------------
    def search_files(
        self,
        pattern:   Pattern = None,
        recursive: bool    = False
    ) -> Generator['File', None, None]:
        """
        Search for files within this folder, optionally recursive, optionally
        matching a regular expression.

        Parameters
        ----------
        pattern : Pattern, optional
            A regular expression to match against file basenames, by default
            None. When omitted, every files in the search folder will be returned.
        recursive : bool, optional
            When True, search recursively into all subfolders of the search
            folder, by default False.

        Yields
        -------
        Folder
            Any files matching the constraints specified.
        """
        for fullpath in get_subfiles(self, pattern=pattern, recursive=recursive):
            yield File(fullpath)

    # --------------------------------------------------------------------------------------------------------------------
    def child(self, name: PathLike) -> Path:
        """
        Return a a child path of this folder created by concatenating the provided
        `name` with this folder's path.

        Parameters
        ----------
        name : PathLike
            A name or relative path of the child.

        Returns
        -------
        Path
            A new `Path` object referring to the child path.
        """
        return Path(self / Path.path_of(name))

    # --------------------------------------------------------------------------------------------------------------------
    def child_file(self, name: PathLike) -> 'File':
        """
        Return a a child file of this folder created by concatenating the provided
        `name` with this folder's path.

        Parameters
        ----------
        name : PathLike
            A name or relative path of the child.

        Returns
        -------
        File
            A new `File` object referring to the child path.
        """
        return File(self / Path.path_of(name))

    # --------------------------------------------------------------------------------------------------------------------
    def child_folder(self, name: PathLike) -> 'Folder':
        """
        Return a a child folder of this folder created by concatenating the provided
        `name` with this folder's path.

        Parameters
        ----------
        name : PathLike
            A name or relative path of the child.

        Returns
        -------
        Folder
            A new `Folder` object referring to the child path.
        """
        return Folder(self / Path.path_of(name))

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def subdirectories(self) -> Generator['Folder', None, None]:
        """
        An iterator over this folder's subdirectories.

        Returns
        -------
        Folder
            Any child folder.
        """
        return self.search_subdirectories(self)

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def files(self) -> PathList:
        """
        An iterator over this folder's files.

        Returns
        -------
        Folder
            Any child file.
        """
        return self.search_files(self)


# =======================================================================================================================
# Class: FolderSet
# =======================================================================================================================
class FolderSet:
    """
    A convenience class to allow attribute/index-style traversal of folder structures.
    """

    def __init__(self, root: PathLike, structure: Optional[Dict[str, Any]] = None):
        """
        Create a new folder set.

        Parameters
        ----------
        root: PathLike
            The path to the folder.
        structure: Optional[Dict[str, Any]], optional
            A hierarchical structure used to limit traversal, by default None
        """
        self._root      = Folder(root)
        self._structure = structure
        self._cache     = dict()

    # --------------------------------------------------------------------------------------------------------------------
    def __truediv__(self, name: PathLike) -> Union['FolderSet', 'File']:
        """
        Convenience method to allow using division operator to access child objects.

        Parameters
        ----------
        name: PathLike
            The relative path to the child.

        Returns
        -------
        Union['FolderSet', 'File']
            Either a file or folder set object depending on what the child path refers to.
        """
        return self[name]

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def folder(self) -> Folder:
        """
        The root folder object of this folder set.

        Returns
        -------
        Folder
            The `Folder` referring to our root path.
        """
        return self._root

    # --------------------------------------------------------------------------------------------------------------------
    def __getattr__(self, name: str) -> Union['FolderSet', 'File']:
        """
        Access a child file or folder with attribute-style access of this folder object.

        Returns
        -------
        Union['FolderSet', 'File']
            Either a file or folder set object depending on what the child path refers to.

        Raises
        ------
        AttributeError
            When no such child file or folder exists.
        """
        try:
            return self[name]
        except Exception as e:
            raise AttributeError(f"Object of type {type(self)} has no '{name}' attribute.") from e

    # --------------------------------------------------------------------------------------------------------------------
    def __getitem__(self, name: str) -> Union['FolderSet', 'File']:
        """
        Access a child file or folder with index-style access of this folder object.

        Returns
        -------
        Union['FolderSet', 'File']
            Either a file or folder set object depending on what the child path refers to.

        Raises
        ------
        KeyError
            When no such child file or folder exists.
        """
        if name.startswith('$$'):
            name = name[2:]

        child = self._root.child(name)
        if not child.exists():
            raise KeyError(f"Object '{name}' is not found in directory '{self._root}'.")

        if child.is_dir():
            if self._structure is not None:
                if name not in self._structure:
                    if (
                        ('$$folders' not in self._structure) or
                        (name not in self._structure['$$folders'])
                    ):
                        raise KeyError(
                            f"Object '{name}' is not defined in the "
                            f"FolderSet structure for '{self._root}', and "
                            "so may not be accessed."
                        )
                    return FolderSet(root=child, structure=None)
                return FolderSet(root=child, structure=self._structure[name])
            return FolderSet(root=child, structure=None)

        elif child.is_file():
            if (self._structure is not None) and ('$$files' in self._structure):
                if name not in self._structure['$$files']:
                    raise KeyError(
                        f"Object '{name}' is not defined in the "
                        f"FolderSet structure for '{self._root}', and "
                        "so may not be accessed."
                    )
            return File(path=child)

        raise KeyError(f"Object '{name}' found in directory '{self._root}' is neither a file nor a directory.")



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os, sys, json, shutil, datetime, configparser, re, subprocess

root_dir = os.path.abspath(
    os.path.join(
        '..',
        '..'
    )
)

sys.path.insert(0, os.path.join(root_dir, 'src' ) )


def get_setup_cfg():
    cfg_filename = os.path.join(root_dir, 'setup.cfg')
    with open(cfg_filename, 'r') as filp:
        cfg = configparser.ConfigParser()
        cfg.read_file(filp, source=cfg_filename)
        return dict(
            ( section, dict( cfg.items(section) ) )
            for section
            in  cfg.sections()
        )

# -- Project information -----------------------------------------------------
setup_cfg = get_setup_cfg()
now       = datetime.datetime.utcnow()
metadata  = setup_cfg["metadata"]
project   = metadata["name"]
author    = metadata["author"]
copyright = "%s, %s" % (now.year, metadata["author"])
version   = u''
release   = u''

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',

    #'sphinxcontrib.apidoc', < tricky getting this to do what i want.>
    'sphinxcontrib.openapi',
    #'sphinxcontrib.blockdiag', < doesn't seem to work anymore
    'sphinx_automodapi.automodapi',
    'numpydoc'
]

#apidoc_module_dir     = '../../src/pt/ptutils'
#apidoc_excluded_paths = []
#apidoc_output_dir     = 'api'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['.templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['.static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'ptutilsdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'ptutils.tex', u'Principled Technologies Boilerplate Utilities Documentation',
     u'Christopher Wolfe', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'ptutils', u'Principled Technologies Boilerplate Utilities Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'ptutils', u'Principled Technologies Boilerplate Utilities Documentation',
     author, 'ptutils', 'A repository of oft used functions and utility classes.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

#===========================================================================
# Options for sphinx_rtd_theme
#===========================================================================
# https://github.com/rtfd/sphinx_rtd_theme
# https://sphinx-rtd-theme.readthedocs.io/en/latest/configuring.html
# https://docs.readthedocs.io/en/latest/guides/adding-custom-css.html
html_theme = "sphinx_rtd_theme"
html_logo  = '.static/img/logo.png'
html_theme_options = {
    'canonical_url':                '',
    'logo_only':                    False,
    'display_version':              True,
    'prev_next_buttons_location':   'bottom',
    'style_external_links':         True,
    # Toc options
    'collapse_navigation':          True,
    'sticky_navigation':            True,
    'navigation_depth':             4,
    'includehidden':                True,
    'titles_only':                  False
}

#===========================================================================
# Options for automodapi
#===========================================================================
# https://sphinx-automodapi.readthedocs.io/en/latest/automodapi.html
automodapi_inheritance_diagram = False
automodapi_toctreedirnm        = 'api'
automodapi_writereprocessed    = False
automodsumm_inherited_members  = True
autosummary_generate           = True
# autosummary_imported_members   = True

#===========================================================================
# Options for NumPyDoc
#===========================================================================
# Note: this is required to be false by automodapi to prevent duplicating
# documentation.
numpydoc_show_class_members    = False
suppress_warnings = ['toc.secnum']

############################################################################
# Utility functions
############################################################################
from json import JSONEncoder

def default_json_encoder(obj):
    try:
        return obj.__dict__
    except:
        pass
    try:
        return magic(obj)
    except:
        pass

    return repr(obj)

def pretty(obj):
    """Return a human-friendly JSON representation of an object (must be JSON-serializable)."""
    return json.dumps(obj, indent=2, sort_keys=True, default=default_json_encoder)

def have_dot():
    """Test if GraphViz's 'dot' program is installed."""
    rv = shutil.which("dot") is not None
    if not rv:
        print("NOTICE: Graphviz does not appear to be installed (could not find 'dot' executable), inheritance diagram generation is disabled.")
    return rv

def native_path(upath):
    return upath.replace('/', os.path.sep)

#===========================================================================
# plug-in to add template processing to rst documents
#===========================================================================
# from https://www.ericholscher.com/blog/2016/jul/25/integrating-jinja-rst-sphinx/
def source_template_preprocessor(app, docname, source):
    """
    Render our pages as a jinja template for fancy templating goodness.
    """
    # Make sure we're outputting HTML
    if app.builder.format != 'html':
        return

    if docname.startswith('api'+os.sep):
        return

    source[0] = app.builder.templates.render_string(
        source[0], vars(app.config)
    )


#===========================================================================
# Main set-up function
#===========================================================================
def setup(app):

    # make custom styles available
    app.add_css_file('css/pt-sphinx-rtd-theme.css')

    # connect the template preprocessor
    app.connect("source-read", source_template_preprocessor)

    # adjust dynamic options
    app.config.automodapi_inheritance_diagram = have_dot()

    try:
        repository_url = subprocess.check_output(['git', 'config', 'remote.origin.url']).strip().decode('utf-8')
        m = re.match(r'^(?:.*/)?([^/]+)', repository_url)
        if m:
            repository_name = m.group(1)
            if repository_name.endswith('.git'):
                repository_name = repository_name[:-4]

        app.config.repository_url = repository_url
        app.config.repository_name = repository_name

    except Exception as e:
        print("WARNING: unable to determine GIT repository URL: %s" %e)

    have_openapi     = False
    openapi_fullpath = None
    openapi_relpath  = None
    try:
        openapi_file = setup_cfg.get("miscellaneous", dict()).get("openapi_file", None)
        if openapi_file:
            fullpath = os.path.join(root_dir, native_path(openapi_file))
            have_openapi = os.path.exists( fullpath )
            if have_openapi:
                openapi_fullpath = fullpath
                openapi_relpath  = os.path.relpath(openapi_fullpath, os.path.abspath(os.path.join('.', 'doc', 'source' )))
    except:
        pass
    print("openapi_fullpath - %s" % openapi_fullpath)
    print("openapi_relpath  - %s" % openapi_relpath)
    print("have_openapi     - %s" % have_openapi)
    app.config.have_openapi     = have_openapi
    app.config.openapi_fullpath = openapi_fullpath
    app.config.openapi_relpath  = openapi_relpath



    app.config.docs_build_date  = now
    app.config.setup_cfg        = setup_cfg
    app.config.author           = setup_cfg["metadata"]["author"]
    app.config.copyright        = u"%s, %s" % (now.year, app.config.author)






#print(globals())
